from __future__ import absolute_import
from six.moves import range
from six.moves import zip
def parse_int(it):
    return int(next(it))
def parse_float(it):
    return float(next(it))
class ListParser:
    def __init__(self, len_parser, item_parser):
        self.len_parser = len_parser
        self.item_parser = item_parser

    def __call__(self, it):
        return [self.item_parser(it) for i in range(self.len_parser(it))]

def make_parser(it):
    tp = next(it)
    if tp == "list":
        len_parser = make_parser(it)
        item_parser = make_parser(it)
        return ListParser(len_parser, item_parser)
    elif tp in "float double float32 float64".split():
        return parse_float
    elif tp in "char uchar short ushort int uint int8 uint8 int16 uint16 int32".split():
        return parse_int
    else:
        raise ValueError("unknown type '%s'" % tp)

def parse_ply(name):
    lines = [l.strip().lower() for l in open(name).readlines()]

    assert lines[0] == "ply"
    assert lines[1].split() == ["format", "ascii", "1.0"]
    
    i = 2

    data_queue = []

    # parse header
    while lines[i] != "end_header":
        words = lines[i].split()
        if words[0] == "element":
            i += 1
            props = []
            lsplit = lines[i].split()
            while lsplit[0] == "property":
                props.append((lsplit[-1], make_parser(iter(lsplit[1:-1]))))
                i += 1
                lsplit = lines[i].split()

            data_queue.append((words[1], int(words[2]), props))
        elif words[0] in ["comment", "created"]:
            i += 1
        else:
            raise ValueError("invalid header field")
    i += 1 # skip end_header

    result = {}

    def parse_line(parsers, line):
        it = iter(line.split())
        result = []
        for p in parsers:
            result.append(p(it))
        return result

    from pytools import Record
    class DataBlock(Record):
        pass

    for name, line_count, props in data_queue:
        prop_names, parsers = list(zip(*props))
        result[name] = DataBlock(
                properties=prop_names, 
                data=[parse_line(parsers, l) for l in lines[i:i+line_count]])

        i += line_count

    return result





from __future__ import division, absolute_import, print_function

__copyright__ = "Copyright (C) 2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from six.moves import range


# {{{ triangle

def test_triangle_refine():
    import meshpy.triangle as triangle
    import math

    segments = 50

    points = [(1, 0), (1, 1), (-1, 1), (-1, -1), (1, -1)]
    n_outer_points = len(points)

    for i in range(0, segments):
        angle = i * 2 * math.pi / segments
        points.append((0.5 * math.cos(angle), 0.5 * math.sin(angle)))

    def round_trip_connect(start, end):
        result = []
        for i in range(start, end):
            result.append((i, i+1))
        result.append((end, start))
        return result

    def needs_refinement(vertices, area):
        vert_origin, vert_destination, vert_apex = vertices
        bary_x = (vert_origin.x + vert_destination.x + vert_apex.x) / 3
        bary_y = (vert_origin.y + vert_destination.y + vert_apex.y) / 3

        dist_center = math.sqrt(bary_x**2 + bary_y**2)
        max_area = 100*(math.fabs(0.002 * (dist_center-0.3)) + 0.0001)
        return area > max_area

    info = triangle.MeshInfo()
    info.set_points(points)
    info.set_holes([(0, 0)])
    info.set_facets(
        round_trip_connect(0, n_outer_points-1)
        +
        round_trip_connect(n_outer_points, len(points)-1))

    mesh = triangle.build(info, refinement_func=needs_refinement)

    triangle.write_gnuplot_mesh("triangles-unrefined.dat", mesh)

    mesh.element_volumes.setup()

    for i in range(len(mesh.elements)):
        mesh.element_volumes[i] = -1
    for i in range(0, len(mesh.elements), 10):
        mesh.element_volumes[i] = 1e-8

    mesh = triangle.refine(mesh)

# }}}


# {{{ tetgen

def test_tetgen():
    from meshpy.tet import MeshInfo, build
    mesh_info = MeshInfo()

    mesh_info.set_points([
        (0, 0, 0),
        (2, 0, 0),
        (2, 2, 0),
        (0, 2, 0),
        (0, 0, 12),
        (2, 0, 12),
        (2, 2, 12),
        (0, 2, 12),
        ])

    mesh_info.set_facets([
        [0, 1, 2, 3],
        [4, 5, 6, 7],
        [0, 4, 5, 1],
        [1, 5, 6, 2],
        [2, 6, 7, 3],
        [3, 7, 4, 0],
        ])

    build(mesh_info)


def test_torus():
    from math import pi, cos, sin
    from meshpy.tet import MeshInfo, build
    from meshpy.geometry import generate_surface_of_revolution, \
            EXT_CLOSED_IN_RZ, GeometryBuilder

    big_r = 3
    little_r = 2.9

    points = 50
    dphi = 2*pi/points

    rz = [(big_r+little_r*cos(i*dphi), little_r*sin(i*dphi))
            for i in range(points)]

    geob = GeometryBuilder()
    geob.add_geometry(*generate_surface_of_revolution(rz,
            closure=EXT_CLOSED_IN_RZ, radial_subdiv=20))

    mesh_info = MeshInfo()
    geob.set(mesh_info)

    build(mesh_info)


def test_tetgen_points():
    from meshpy.tet import MeshInfo, build, Options

    import numpy as np
    points = np.random.randn(10000, 3)

    mesh_info = MeshInfo()
    mesh_info.set_points(points)
    options = Options("")
    mesh = build(mesh_info, options=options)

    print(len(mesh.points))
    print(len(mesh.elements))

    #mesh.write_vtk("test.vtk")

# }}}


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from pytest import main
        main([__file__])

# vim: foldmethod=marker



from __future__ import division, absolute_import

__doc__ = """

Geometry builder
----------------

.. autoclass:: GeometryBuilder

Geometries
----------

These functions are designed so that their output can be splat-passed to
:meth:`GeometryBuilder.add_geometry`::

    builder = GeometryBuilder()
    builder.add_geometry(*make_ball(10))

.. autoclass:: Marker
    :members:
    :undoc-members:

.. autofunction:: make_box
.. autofunction:: make_circle
.. autofunction:: make_ball
.. autofunction:: make_cylinder

Extrusions and surfaces of revolution
-------------------------------------

.. data:: EXT_OPEN
.. data:: EXT_CLOSED_IN_RZ

.. autofunction:: generate_extrusion
.. autofunction:: generate_surface_of_revolution
"""

import numpy as np
from six.moves import range
from six.moves import zip


# {{{ geometry building

def bounding_box(points):
    return (
            np.asarray(np.min(points, axis=0), dtype=np.float64),
            np.asarray(np.max(points, axis=0), dtype=np.float64))


def is_multi_polygon(facets):
    if not len(facets):
        return False

    try:
        facets[0][0][0]  # facet 0, poly 0, point 0
    except TypeError:
        # pure python raises this
        return False
    except IndexError:
        # numpy raises this
        return False
    else:
        return True


def offset_point_indices(facets, offset):
    if is_multi_polygon(facets):
        return [[tuple(p_i+offset for p_i in poly)
            for poly in facet]
            for facet in facets]
    else:
        return [tuple(p_i+offset for p_i in facet) for facet in facets]


class GeometryBuilder(object):
    """
    .. automethod:: add_geometry
    .. automethod:: set
    .. automethod:: wrap_in_box
    .. automethod:: bounding_box
    .. automethod:: center
    .. automethod:: apply_transform
    """
    def __init__(self):
        self.points = []
        self.facets = []
        self.facet_hole_starts = None
        self.facet_markers = None
        self.point_markers = None

    def add_geometry(self, points, facets, facet_hole_starts=None,
            facet_markers=None, point_markers=None):
        if isinstance(facet_markers, int):
            facet_markers = len(facets) * [facet_markers]

        if facet_hole_starts and not self.facet_hole_starts:
            self.facet_hole_starts = len(self.facets) * []
        if facet_markers and not self.facet_markers:
            self.facet_markers = len(self.facets) * [0]
        if point_markers and not self.point_markers:
            self.point_markers = len(self.points) * [0]

        if not facet_hole_starts and self.facet_hole_starts:
            facet_hole_starts = len(facets) * [[]]
        if not facet_markers and self.facet_markers:
            facet_markers = len(facets) * [0]
        if not point_markers and self.point_markers:
            point_markers = len(points) * [0]

        if is_multi_polygon(facets) and not is_multi_polygon(self.facets):
            self.facets = [[facet] for facet in self.facets]

        if not is_multi_polygon(facets) and is_multi_polygon(self.facets):
            facets = [[facet] for facet in facets]

        self.facets.extend(offset_point_indices(facets, len(self.points)))
        self.points.extend(points)

        if facet_markers:
            self.facet_markers.extend(facet_markers)
            assert len(facets) == len(facet_markers)
        if facet_hole_starts:
            self.facet_hole_starts.extend(facet_hole_starts)
            assert len(facets) == len(facet_hole_starts)
        if point_markers:
            self.point_markers.extend(point_markers)
            assert len(points) == len(point_markers)

    def add_cycle(self, points, facet_markers=None, point_markers=None):
        def make_facets():
            end = len(points)-1
            for i in range(end):
                yield i, i+1
            yield end, 0

        self.add_geometry(points, list(make_facets()),
                facet_markers=facet_markers,
                point_markers=point_markers)

    def dimensions(self):
        return len(self.points[0])

    def set(self, mesh_info):
        """Transfer the built geometry into a :class:`meshpy.triangle.MeshInfo`
        or a :class:`meshpy.tet.MeshInfo`.
        """

        mesh_info.set_points(self.points, self.point_markers)
        if self.facet_hole_starts or is_multi_polygon(self.facets):
            mesh_info.set_facets_ex(self.facets,
                    self.facet_hole_starts, self.facet_markers)
        else:
            mesh_info.set_facets(self.facets, self.facet_markers)

    def mesher_module(self):
        dim = self.dimensions()
        if dim == 2:
            import meshpy.triangle
            return meshpy.triangle
        elif dim == 3:
            import meshpy.tet
            return meshpy.tet
        else:
            raise ValueError("unsupported dimensionality %d" % dim)

    def bounding_box(self):
        return bounding_box(self.points)

    def center(self):
        a, b = bounding_box(self.points)
        return (a+b)/2

    def wrap_in_box(self, distance, subdivisions=None):
        """
        :param subdivisions: is a tuple of integers specifying
          the number of subdivisions along each axis.
        """

        a, b = bounding_box(self.points)
        points, facets, _, facet_markers = \
                make_box(a-distance, b+distance, subdivisions)

        self.add_geometry(points, facets, facet_markers=facet_markers)

    def apply_transform(self, f):
        self.points = [f(x) for x in self.points]

# }}}


# {{{ actual geometries

class Marker:
    MINUS_X = 1
    PLUS_X = 2
    MINUS_Y = 3
    PLUS_Y = 4
    MINUS_Z = 5
    PLUS_Z = 6
    SHELL = 100

    FIRST_USER_MARKER = 1000


def make_box(a, b, subdivisions=None):
    """
    :param subdivisions: is a tuple of integers specifying
      the number of subdivisions along each axis.
    """

    a = [float(ai) for ai in a]
    b = [float(bi) for bi in b]

    assert len(a) == len(b)

    dimensions = len(a)
    if dimensions == 2:
        # CAUTION: Do not change point or facet order here.
        # Other code depends on this staying the way it is.

        points = [
                (a[0], a[1]),
                (b[0], a[1]),
                (b[0], b[1]),
                (a[0], b[1]),
                ]

        facets = [(0, 1), (1, 2), (2, 3), (3, 0)]

        facet_markers = [
                Marker.MINUS_Y, Marker.PLUS_X,
                Marker.PLUS_Y, Marker.MINUS_X]

    elif dimensions == 3:
        #    7--------6
        #   /|       /|
        #  4--------5 |  z
        #  | |      | |  ^
        #  | 3------|-2  | y
        #  |/       |/   |/
        #  0--------1    +--->x

        points = [
                (a[0], a[1], a[2]),
                (b[0], a[1], a[2]),
                (b[0], b[1], a[2]),
                (a[0], b[1], a[2]),
                (a[0], a[1], b[2]),
                (b[0], a[1], b[2]),
                (b[0], b[1], b[2]),
                (a[0], b[1], b[2]),
                ]

        facets = [
                (0, 1, 2, 3),
                (0, 1, 5, 4),
                (1, 2, 6, 5),
                (7, 6, 2, 3),
                (7, 3, 0, 4),
                (4, 5, 6, 7)
                ]

        facet_markers = [Marker.MINUS_Z, Marker.MINUS_Y, Marker.PLUS_X,
                Marker.PLUS_Y, Marker.MINUS_X, Marker.PLUS_Z]
    else:
        raise ValueError("unsupported dimension count: %d" % len(a))

    if subdivisions is not None:
        if dimensions != 2:
            raise NotImplementedError(
                    "subdivision not implemented for any "
                    "dimension count other than 2")

        from meshpy.triangle import subdivide_facets
        points, facets, facet_markers = subdivide_facets(
                [subdivisions[0], subdivisions[1],
                    subdivisions[0], subdivisions[1]],
                points, facets, facet_markers)

    return points, facets, None, facet_markers


def make_circle(r, center=(0, 0), subdivisions=40, marker=Marker.SHELL):
    def round_trip_connect(seq):
        result = []
        for i in range(len(seq)):
            result.append((i, (i+1) % len(seq)))
        return result

    phi = np.linspace(0, 2*np.pi, num=subdivisions, endpoint=False)
    cx, cy = center
    x = r*np.cos(phi) + cx
    y = r*np.sin(phi) + cy

    return ([np.array(pt) for pt in zip(x, y)],
            round_trip_connect(list(range(subdivisions))),
            None,
            subdivisions*[marker])


def make_ball(r, subdivisions=10):
    from math import pi, cos, sin

    dphi = pi/subdivisions

    def truncate(my_r):
        if abs(my_r) < 1e-9*r:
            return 0
        else:
            return my_r

    rz = [(truncate(r*sin(i*dphi)), r*cos(i*dphi)) for i in range(subdivisions+1)]

    return generate_surface_of_revolution(
            rz, closure=EXT_OPEN, radial_subdiv=subdivisions)


def make_cylinder(radius, height, radial_subdivisions=10,
        height_subdivisions=1):
    dz = height/height_subdivisions
    rz = [(0, 0)] \
            + [(radius, i*dz) for i in range(height_subdivisions+1)] \
            + [(0, height)]
    ring_markers = [Marker.MINUS_Z] \
            + ((height_subdivisions)*[Marker.SHELL]) \
            + [Marker.PLUS_Z]

    return generate_surface_of_revolution(rz,
            closure=EXT_OPEN, radial_subdiv=radial_subdivisions,
            ring_markers=ring_markers)

# }}}


# {{{ extrusions

def _is_same_float(a, b, threshold=1e-10):
    if abs(a) > abs(b):
        a, b = b, a

    # now abs(a) <= abs(b) always
    return abs(b) < threshold or abs(a-b) < threshold*abs(b)


EXT_OPEN = 0
EXT_CLOSED_IN_RZ = 1


def generate_extrusion(rz_points, base_shape, closure=EXT_OPEN,
        point_idx_offset=0, ring_point_indices=None,
        ring_markers=None, rz_closure_marker=0):
    """Extrude a given connected *base_shape* (a list of (x,y) points)
    along the z axis. For each step in the extrusion, the base shape
    is multiplied by a radius and shifted in the z direction. Radius
    and z offset are given by *rz_points*, which is a list of
    (r, z) tuples.

    Returns ``(points, facets, facet_holestarts, markers)``, where *points* is
    a list of (3D) points and facets is a list of polygons. Each polygon is, in
    turn, represented by a tuple of indices into *points*. If
    *point_idx_offset* is not zero, these indices start at that number.
    *markers* is a list equal in length to *facets*, each specifying the facet
    marker of that facet.  *facet_holestarts* is also equal in length to
    *facets*, each element is a list of hole starting points for the
    corresponding facet.

    Use L{MeshInfo.set_facets_ex} to add the extrusion to a L{MeshInfo}
    structure.

    The extrusion proceeds by generating quadrilaterals connecting each
    ring.  If any given radius in *rz_points* is 0, triangle fans are
    produced instead of quads to provide non-degenerate closure.

    If *closure* is :data:`EXT_OPEN`, no efforts are made to put end caps on the
    extrusion.

    If *closure* is :data:`EXT_CLOSED_IN_RZ`, then a torus-like structure
    is assumed and the last ring is just connected to the first.

    If *ring_markers* is not None, it is an list of markers added to each
    ring. There should be len(rz_points)-1 entries in this list.
    If rings are added because of closure options, they receive the
    corresponding *XXX_closure_marker*.  If *facet_markers* is given, this function
    returns (points, facets, markers), where markers is is a list containing
    a marker for each generated facet. Unspecified markers generally
    default to 0.

    If *ring_point_indices* is given, it must be a list of the same
    length as *rz_points*. Each entry in the list may either be None,
    or a list of point indices. This list must contain the same number
    of points as the *base_shape*; it is taken as the indices of
    pre-existing points that are to be used for the given ring, instead
    of generating new points.
    """

    assert len(rz_points) > 0

    if ring_markers is not None:
        assert len(rz_points) == len(ring_markers)+1

    def get_ring(ring_idx):
        try:
            return rings[ring_idx]
        except KeyError:
            # need to generate fresh ring, continue
            pass

        p_indices = None
        if ring_point_indices is not None:
            p_indices = ring_point_indices[ring_idx]

        first_idx = point_idx_offset+len(points)

        r, z = rz_points[ring_idx]

        if r == 0:
            p_indices = (first_idx,)
            points.append((0, 0, z))
        else:
            p_indices = tuple(range(first_idx, first_idx+len(base_shape)))
            points.extend([(x*r, y*r, z) for (x, y) in base_shape])

        rings[ring_idx] = p_indices
        return p_indices

    def pair_with_successor(l):
        n = len(l)
        return [(l[i], l[(i+1) % n]) for i in range(n)]

    def add_polygons(new_polys, marker):
        """Add several new facets, each polygon in new_polys corresponding
        to a new facet.
        """
        facets.extend([poly] for poly in new_polys)
        markers.extend(len(new_polys)*[marker])
        holelists.extend(len(new_polys)*[[]])

    def add_facet(facet_polygons, holestarts, marker):
        """Add a single facet, with each polygon in *facet_polygons*
        belonging to a single facet.
        """
        facets.append(facet_polygons)
        markers.append(marker)
        holelists.append(holestarts)

    def connect_ring(ring1_idx, ring2_idx, marker):
        r1, z1 = rz_points[ring1_idx]
        r2, z2 = rz_points[ring2_idx]

        if _is_same_float(z2, z1):
            assert not _is_same_float(r1, r2)
            # we're moving purely outward--this doesn't need fans, only plane
            # surfaces. Special casing this leads to more freedom for TetGen
            # and hence better meshes.

            if r1 == 0:
                # make opening surface
                if r2 != 0:
                    add_polygons([get_ring(ring2_idx)], marker=marker)
            elif r2 == 0:
                # make closing surface
                add_polygons([get_ring(ring1_idx)], marker=marker)
            else:
                # make single-surface interface with hole
                add_facet([
                    get_ring(ring1_idx),
                    get_ring(ring2_idx),
                    ],
                    holestarts=[(0, 0, z1)], marker=marker)
        else:
            ring1 = get_ring(ring1_idx)
            ring2 = get_ring(ring2_idx)
            if r1 == 0:
                # make opening fan
                assert len(ring1) == 1
                start_pt = ring1[0]

                if r2 != 0:
                    add_polygons(
                            [(start_pt, succ, pt)
                            for pt, succ in pair_with_successor(ring2)],
                            marker=marker)
            elif r2 == 0:
                # make closing fan
                assert len(ring2) == 1
                end_pt = ring2[0]
                add_polygons(
                        [(pt, succ, end_pt)
                        for pt, succ in pair_with_successor(ring1)],
                        marker=marker)
            else:
                # make quad strip
                pairs1 = pair_with_successor(ring1)
                pairs2 = pair_with_successor(ring2)
                add_polygons(
                        [(a, b, c, d) for ((a, b), (d, c)) in zip(pairs1, pairs2)],
                        marker=marker)

    points = []
    facets = []
    markers = []
    holelists = []

    rings = {}

    # pre-populate ring dict with ring_point_indices
    if ring_point_indices is not None:
        for i, ring_points in enumerate(ring_point_indices):
            if ring_points is not None:
                assert isinstance(ring_points, tuple)

                if rz_points[i][0] == 0:
                    assert len(ring_points) == 1
                else:
                    assert len(ring_points) == len(base_shape), \
                            ("Ring points length (%d) does not "
                                    "match base shape length (%d)"
                                    % (len(ring_points), len(base_shape)))

                rings[i] = ring_points

    for i in range(len(rz_points)-1):
        if ring_markers is not None:
            ring_marker = ring_markers[i]
        else:
            ring_marker = 0

        connect_ring(i, i+1, ring_marker)

    if closure == EXT_CLOSED_IN_RZ:
        connect_ring(len(rz_points)-1, 0, rz_closure_marker)

    return points, facets, holelists, markers


def generate_surface_of_revolution(rz_points,
        closure=EXT_OPEN, radial_subdiv=16,
        point_idx_offset=0, ring_point_indices=None,
        ring_markers=None, rz_closure_marker=0):
    from math import sin, cos, pi

    dphi = 2*pi/radial_subdiv
    base_shape = [(cos(dphi*i), sin(dphi*i)) for i in range(radial_subdiv)]
    return generate_extrusion(
            rz_points, base_shape, closure=closure,
            point_idx_offset=point_idx_offset,
            ring_point_indices=ring_point_indices,
            ring_markers=ring_markers, rz_closure_marker=rz_closure_marker,
            )

# }}}

# vim: foldmethod=marker



from __future__ import absolute_import
from __future__ import print_function
from six.moves import range
from six.moves import zip
class _Table:
    def __init__(self):
        self.Rows = []

    def add_row(self, row):
        self.Rows.append([str(i) for i in row])

    def __str__(self):
        columns = len(self.Rows[0])
        col_widths = [max(len(row[i]) for row in self.Rows)
                      for i in range(columns)]

        lines = [
            " ".join([cell.ljust(col_width)
                      for cell, col_width in zip(row, col_widths)])
            for row in self.Rows]
        return "\n".join(lines)




def _linebreak_list(list, per_line=10, pad=None):
    def format(s):
        if pad is None:
            return str(s)
        else:
            return str(s).rjust(pad)

    result = ""
    while len(list) > per_line:
        result += " ".join(format(l) for l in list[:per_line]) + "\n"
        list = list[per_line:]
    return result + " ".join(format(l) for l in list)



class MeshInfoBase:
    @property
    def face_vertex_indices_to_face_marker(self):
        try:
            return self._fvi2fm
        except AttributeError:
            result = {}

            for i, face in enumerate(self.faces):
                result[frozenset(face)] = self.face_markers[i]

            self._fvi2fm = result
            return result





    def set_points(self, points, point_markers=None):
        if point_markers is not None:
            assert len(point_markers) == len(point_markers)

        self.points.resize(len(points))

        for i, pt in enumerate(points):
            self.points[i] = pt

        if point_markers is not None:
            self.point_markers.setup()
            for i, mark in enumerate(point_markers):
                self.point_markers[i] = mark





    def set_holes(self, hole_starts):
        self.holes.resize(len(hole_starts))
        for i, hole in enumerate(hole_starts):
            self.holes[i] = hole




    def write_neu(self, outfile, bc={}, periodicity=None, description="MeshPy Output"):
        """Write the mesh out in (an approximation to) Gambit neutral mesh format.

        outfile is a file-like object opened for writing.

        bc is a dictionary mapping single face markers (or frozensets of them)
        to a tuple (bc_name, bc_code).

        periodicity is either a tuple (face_marker, (px,py,..)) giving the
        face marker of the periodic boundary and the period in each coordinate
        direction (0 if none) or the value None for no periodicity.
        """

        from meshpy import version
        from datetime import datetime

        # header --------------------------------------------------------------
        outfile.write("CONTROL INFO 2.1.2\n")
        outfile.write("** GAMBIT NEUTRAL FILE\n")
        outfile.write("%s\n" % description)
        outfile.write("PROGRAM: MeshPy VERSION: %s\n" % version)
        outfile.write("%s\n" % datetime.now().ctime())

        bc_markers = list(bc.keys())
        if periodicity:
            periodic_marker, periods = periodicity
            bc_markers.append(periodic_marker)

        assert len(self.points)

        dim = len(self.points[0])
        data = (
                ("NUMNP", len(self.points)),
                ("NELEM", len(self.elements)),
                ("NGRPS", 1),
                ("NBSETS", len(bc_markers)),
                ("NDFCD", dim),
                ("NDFVL", dim),
                )

        tbl = _Table()
        tbl.add_row(key for key, value in data)
        tbl.add_row(value for key, value in data)
        outfile.write(str(tbl))
        outfile.write("\n")
        outfile.write("ENDOFSECTION\n")

        # nodes ---------------------------------------------------------------
        outfile.write("NODAL COORDINATES 2.1.2\n")
        for i, pt in enumerate(self.points):
            outfile.write("%d %s\n" %
                    (i+1, " ".join(repr(c) for c in pt)))
        outfile.write("ENDOFSECTION\n")

        # elements ------------------------------------------------------------
        outfile.write("ELEMENTS/CELLS 2.1.2\n")
        if dim == 2:
            eltype = 3
        else:
            eltype = 6

        for i, el in enumerate(self.elements):
            outfile.write("%8d%3d%3d %s\n" %
                    (i+1, eltype, len(el),
                        "".join("%8d" % (p+1) for p in el)))
        outfile.write("ENDOFSECTION\n")

        # element groups ------------------------------------------------------
        outfile.write("ELEMENT GROUP 1.3.0\n")
        # FIXME
        i = 0
        grp_elements = list(range(len(self.elements)))
        material = 1
        flags = 0
        outfile.write("GROUP:%11d ELEMENTS:%11d MATERIAL:%11s NFLAGS: %11d\n"
                % (1, len(grp_elements), repr(material), flags))
        outfile.write(("epsilon: %s\n" % material).rjust(32)) # FIXME
        outfile.write("0\n")
        outfile.write(_linebreak_list([str(i+1) for i in grp_elements],
            pad=8)
                + "\n")
        outfile.write("ENDOFSECTION\n")

        # boundary conditions -------------------------------------------------
        # build mapping face -> (tet, neu_face_index)
        face2el = {}

        if dim == 2:
            for ti, el in enumerate(self.elements):
                # Sledge++ Users' Guide, figure 4
                faces = [
                        frozenset([el[0], el[1]]),
                        frozenset([el[1], el[2]]),
                        frozenset([el[2], el[0]]),
                        ]
                for fi, face in enumerate(faces):
                    face2el.setdefault(face, []).append((ti, fi+1))

        elif dim == 3:
            face2el = {}
            for ti, el in enumerate(self.elements):
                # Sledge++ Users' Guide, figure 5
                faces = [
                        frozenset([el[1], el[0], el[2]]),
                        frozenset([el[0], el[1], el[3]]),
                        frozenset([el[1], el[2], el[3]]),
                        frozenset([el[2], el[0], el[3]]),
                        ]
                for fi, face in enumerate(faces):
                    face2el.setdefault(face, []).append((ti, fi+1))

        else:
            raise ValueError("invalid number of dimensions (%d)" % dim)

        # actually output bc sections
        if not self.faces.allocated:
            from warnings import warn
            warn("no exterior faces in mesh data structure, not writing boundary conditions")
        else:
            # requires -f option in tetgen, -e in triangle

            for bc_marker in bc_markers:
                if isinstance(bc_marker, frozenset):
                    face_indices = [i
                            for i, face in enumerate(self.faces)
                            if self.face_markers[i] in bc_marker]
                else:
                    face_indices = [i
                            for i, face in enumerate(self.faces)
                            if bc_marker == self.face_markers[i]]

                if not face_indices:
                    continue

                outfile.write("BOUNDARY CONDITIONS 2.1.2\n")
                if bc_marker in bc:
                    # regular BC

                    bc_name, bc_code = bc[bc_marker]
                    outfile.write("%32s%8d%8d%8d%8d\n"
                            % (bc_name,
                                1, # face BC
                                len(face_indices),
                                0, # zero additional values per face,
                                bc_code,
                                )
                            )
                else:
                    # periodic BC

                    outfile.write("%s%s%8d%8d%8d\n"
                            % ("periodic", " ".join(repr(p) for p in periods),
                                len(face_indices),
                                0, # zero additional values per face,
                                0,
                                )
                            )

                for i, fi in enumerate(face_indices):
                    face_nodes = frozenset(self.faces[fi])
                    adj_el = face2el[face_nodes]
                    assert len(adj_el) == 1

                    el_index, el_face_number = adj_el[0]

                    outfile.write("%10d%5d%5d\n" %
                            (el_index+1, eltype, el_face_number))

                outfile.write("ENDOFSECTION\n")

            outfile.close()
            # FIXME curved boundaries?
            # FIXME proper element group support





def dump_array(name, array):
    print("array %s: %d elements, %d values per element" % (name, len(array), array.unit))

    if len(array) == 0 or array.unit == 0:
        return

    try:
        array[0]
    except RuntimeError:
        print("  not allocated")
        return

    for i, entry in enumerate(array):
        if isinstance(entry, list):
            print("  %d: %s" % (i, ",".join(str(sub) for sub in entry)))
        else:
            print("  %d: %s" % (i, entry))




from __future__ import absolute_import

import six
from six.moves import range


def uniform_refine_triangles(points, elements, factor=2):
    new_points = points[:]
    new_elements = []
    old_face_to_new_faces = {}
    face_point_dict = {}

    points_per_edge = factor+1

    def get_refined_face(a, b):
        if a > b:
            a, b = b, a
            flipped = True
        else:
            flipped = False

        try:
            face_points = face_point_dict[a, b]
        except KeyError:
            a_pt, b_pt = [points[idx] for idx in [a, b]]
            dx = (b_pt - a_pt)/factor

            # build subdivided facet
            face_points = [a]

            for i in range(1, points_per_edge-1):
                face_points.append(len(new_points))
                new_points.append(a_pt + dx*i)

            face_points.append(b)

            face_point_dict[a, b] = face_points

            # build old_face_to_new_faces
            old_face_to_new_faces[frozenset([a, b])] = [
                    (face_points[i], face_points[i+1])
                    for i in range(factor)]

        if flipped:
            return face_points[::-1]
        else:
            return face_points

    for a, b, c in elements:
        a_pt, b_pt, c_pt = [points[idx] for idx in [a, b, c]]
        dr = (b_pt - a_pt)/factor
        ds = (c_pt - a_pt)/factor

        ab_refined, bc_refined, ac_refined = [
                get_refined_face(*pt_indices)
                for pt_indices in [(a, b), (b, c), (a, c)]]

        el_point_dict = {}

        # fill out edges of el_point_dict
        for i in range(points_per_edge):
            el_point_dict[i, 0] = ab_refined[i]
            el_point_dict[0, i] = ac_refined[i]
            el_point_dict[points_per_edge-1-i, i] = bc_refined[i]

        # fill out interior of el_point_dict
        for i in range(1, points_per_edge-1):
            for j in range(1, points_per_edge-1-i):
                el_point_dict[i, j] = len(new_points)
                new_points.append(a_pt + dr*i + ds*j)

        # generate elements
        for i in range(0, points_per_edge-1):
            for j in range(0, points_per_edge-1-i):
                new_elements.append((
                    el_point_dict[i, j],
                    el_point_dict[i+1, j],
                    el_point_dict[i, j+1],
                    ))
                if i+1+j+1 <= factor:
                    new_elements.append((
                        el_point_dict[i+1, j+1],
                        el_point_dict[i+1, j],
                        el_point_dict[i, j+1],
                        ))

    from meshpy.triangle import MeshInfo
    mi = MeshInfo()
    mi.set_points(new_points)
    mi.elements.resize(len(new_elements))
    for i, el in enumerate(new_elements):
        mi.elements[i] = el
    from meshpy.triangle import write_gnuplot_mesh
    write_gnuplot_mesh("mesh.dat", mi)

    return new_points, new_elements, old_face_to_new_faces


def make_swizzle_matrix(spec):
    import numpy
    axes = ["x", "y", "z"]

    mapping = dict((axis, axis) for axis in axes)
    for one_spec in spec.split(","):
        import_axis, final_axis = one_spec.split(":")
        mapping[import_axis] = final_axis

    assert set(mapping.keys()) == set(axes), \
            "axis mapping not complete"
    assert set(axis.lstrip("-") for axis in six.itervalues(mapping)) == set(axes), \
            "Axis mapping not onto"

    n = len(axes)
    result = numpy.zeros((n, n), dtype=int)

    for imp_axis, final_axis in six.iteritems(mapping):
        imp_axis = axes.index(imp_axis)

        sign = 1
        while final_axis.startswith("-"):
            sign *= -1
            final_axis = final_axis[1:]
        final_axis = axes.index(final_axis)

        result[final_axis, imp_axis] = sign

    return result



from __future__ import absolute_import
from __future__ import print_function
from meshpy.common import MeshInfoBase, dump_array
import meshpy._internals as internals
import six
from six.moves import range


class MeshInfo(internals.TetMeshInfo, MeshInfoBase):
    def set_facets(self, facets, markers=None):
        """Set a list of simple, single-polygon factes. Unlike :meth:`set_facets_ex`,
        :meth:`set_facets` does not allow hole and only lets you use a single
        polygon per facet.

        :param facets: a list of facets, where each facet is a single
          polygons, represented by a list of point indices.
        :param markers: Either None or a list of integers of the same
          length as *facets*. Each integer is the facet marker assigned
          to its corresponding facet.

        :note: When the above says "list", any repeatable iterable
          also accepted instead.
        """

        if markers:
            assert len(markers) == len(facets)

        self.facets.resize(len(facets))

        for i, vlist in enumerate(facets):
            facet = self.facets[i]
            polys = facet.polygons
            polys.resize(1)
            poly = facet.polygons[0]
            poly.vertices.resize(len(vlist))
            for j, pt_idx in enumerate(vlist):
                poly.vertices[j] = pt_idx

        if markers:
            self.facet_markers.setup()
            for i, mark in enumerate(markers):
                self.facet_markers[i] = mark

    def set_facets_ex(self, facets, facet_holestarts=None, markers=None):
        """Set a list of complicated factes. Unlike :meth:`set_facets`,
        :meth:`set_facets_ex` allows holes and multiple polygons per
        facet.

        :param facets: a list of facets, where each facet is a list
          of polygons, and each polygon is represented by a list
          of point indices.
        :param facet_holestarts: Either None or a list of hole starting points
          for each facet. Each facet may have several hole starting points.
          The mesh generator starts "eating" a hole into the facet at each
          starting point and continues until it hits a polygon specified
          in this facet's record in *facets*.
        :param markers: Either None or a list of integers of the same
          length as *facets*. Each integer is the facet marker assigned
          to its corresponding facet.

        :note: When the above says "list", any repeatable iterable
          also accepted instead.
        """

        if markers:
            assert len(markers) == len(facets)
        if facet_holestarts is not None:
            assert len(facet_holestarts) == len(facets)

        self.facets.resize(len(facets))
        for i_facet, poly_list in enumerate(facets):
            facet = self.facets[i_facet]
            polys = facet.polygons

            polys.resize(len(poly_list))
            for i_poly, vertex_list in enumerate(poly_list):
                poly = facet.polygons[i_poly]

                poly.vertices.resize(len(vertex_list))
                for i_point, point in enumerate(vertex_list):
                    poly.vertices[i_point] = point

            if facet_holestarts is not None:
                hole_list = facet_holestarts[i_facet]
                facet_holes = facet.holes
                facet_holes.resize(len(hole_list))
                for i_hole, hole_start in enumerate(hole_list):
                    for i_coordinate, co_value in enumerate(hole_start):
                        facet_holes[i_hole, i_coordinate] = co_value

        if markers:
            self.facet_markers.setup()
            for i, mark in enumerate(markers):
                self.facet_markers[i] = mark

    def dump(self):
        for name in ["points"]:
            dump_array(name, getattr(self, name))
        for ifacet, facet in enumerate(self.faces):
            print("facet %d:" % ifacet)
            for ipolygon, polygon in enumerate(facet.polygons):
                print("  polygon %d: vertices [%s]" % \
                        (ipolygon, ",".join(str(vi) for vi in polygon.vertices)))

    def write_vtk(self, filename):
        import pyvtk
        vtkelements = pyvtk.VtkData(
            pyvtk.UnstructuredGrid(
                self.points,
                tetra=self.elements),
            "Mesh")
        vtkelements.tofile(filename)

    def set_elements(self, elements):
        self.elements.resize(len(elements))

        for i, element in enumerate(elements):
            self.elements[i] = element

    def set_element_constraints(self, element_constraints):
        self.element_volumes.setup()

        for i in range(len(self.element_volumes)):
            if i in element_constraints:
                self.element_volumes[i] = element_constraints[i]
            else:
                self.element_volumes[i] = -1


class Options(internals.Options):
    def __init__(self, switches, **kwargs):
        internals.Options.__init__(self)
        if len(switches) == 0:
            from warnings import warn
            warn("Recommend non-empty 'switches' for crash-free meshing")
        self.parse_switches(switches)
        self.quiet = 1

        for k, v in six.iteritems(kwargs):
            try:
                getattr(self, k)
            except AttributeError:
                raise ValueError("invalid option: %s" % k)
            else:
                setattr(self, k, v)


def tetrahedralize(mesh_info, options):
    mesh = MeshInfo()

    # restore "C" locale--otherwise tetgen might mis-parse stuff like "a0.01"
    try:
        import locale
    except ImportError:
        have_locale = False
    else:
        have_locale = True
        prev_num_locale = locale.getlocale(locale.LC_NUMERIC)
        locale.setlocale(locale.LC_NUMERIC, "C")

    try:
        internals.tetrahedralize(options, mesh_info, mesh)
    finally:
        # restore previous locale if we've changed it
        if have_locale:
            locale.setlocale(locale.LC_NUMERIC, prev_num_locale)

    return mesh


def build(mesh_info, options=Options("pq"), verbose=False,
        attributes=False, volume_constraints=False, max_volume=None,
        diagnose=False, insert_points=None):
    if not verbose:
        options.quiet = 1

    if insert_points is not None:
        options.insertaddpoints = 1

    if attributes:
        options.regionattrib = 1
    if volume_constraints:
        options.varvolume = 1
    if max_volume:
        options.fixedvolume = 1
        options.maxvolume = max_volume
    if diagnose:
        options.diagnose = 1

    return tetrahedralize(mesh_info, options)



from __future__ import division
from __future__ import absolute_import
from meshpy.common import MeshInfoBase, dump_array
import meshpy._internals as internals
from six.moves import range
from six.moves import zip


class MeshInfo(internals.TriMeshInfo, MeshInfoBase):
    _constituents = [
            "points", "point_attributes", "point_markers",
            "elements", "element_attributes", "element_volumes",
            "neighbors",
            "facets", "facet_markers",
            "holes",
            "regions",
            "faces", "face_markers",
            "normals",
            ]

    def __getstate__(self):
        return self.number_of_point_attributes, \
               self.number_of_element_attributes, \
               [(name, getattr(self, name)) for name in self._constituents]

    def __setstate__(self, xxx_todo_changeme):
        (p_attr_count, e_attr_count, state) = xxx_todo_changeme
        self.number_of_point_attributes = p_attr_count
        self.number_of_element_attributes = e_attr_count
        for name, array in state:
            if name not in self._constituents:
                raise RuntimeError("Unknown constituent during unpickling")

            dest_array = getattr(self, name)

            if array is None:
                dest_array.deallocate()
            else:
                if len(dest_array) != len(array):
                    dest_array.resize(len(array))
                if not dest_array.allocated and len(array) > 0:
                    dest_array.setup()

                for i, tup in enumerate(array):
                    for j, v in enumerate(tup):
                        dest_array[i, j] = v

    def set_facets(self, facets, facet_markers=None):
        self.facets.resize(len(facets))

        for i, facet in enumerate(facets):
            self.facets[i] = facet

        if facet_markers is not None:
            self.facet_markers.setup()
            for i, mark in enumerate(facet_markers):
                self.facet_markers[i] = mark

    def dump(self):
        for name in self._constituents:
            dump_array(name, getattr(self, name))


def subdivide_facets(subdivisions, points, facets, facet_markers=None):
    """Return a new facets array in which the original facets are
    each subdivided into C{subdivisions} subfacets.

    This routine is useful if you have to prohibit the insertion of Steiner
    points on the boundary  of your triangulation to allow the mesh to conform
    either to itself periodically or another given mesh. In this case, you may
    use this routine to create the necessary resolution along the boundary
    in a predefined way.

    @arg subdivisions: Either an C{int}, indicating a uniform number of subdivisions
      throughout, or a list of the same length as C{facets}, specifying a subdivision
      count for each individual facet.
    @arg points: A list of points referred to from the facets list.
    @arg facets: The list of old facets, in the form C{[(p1, p2), (p3,p4), ...]}.
    @arg facet_markers: Either C{None} or a list of facet markers of the same length
      as C{facets}.
    @return: The new tuple C{(new_points, new_facets)}.
      (Or C{(new_points, new_facets, new_facet_markers)} if C{facet_markers} is not
      C{None}.)
    """

    def intermediate_points(pa, pb, n):
        for i in range(1, n):
            tau = i/n
            yield [pai*(1-tau) + tau*pbi for pai, pbi in zip(pa, pb)]

    if isinstance(subdivisions, int):
        from itertools import repeat
        subdiv_it = repeat(subdivisions, len(facets))
    else:
        assert len(facets) == len(subdivisions)
        subdiv_it = subdivisions.__iter__()

    new_points = points[:]
    new_facets = []

    if facet_markers is not None:
        assert len(facets) == len(facet_markers)
        new_facet_markers = []

    for facet_idx, ((pidx_a, pidx_b), subdiv) in enumerate(zip(facets, subdiv_it)):
        facet_points = [pidx_a]
        for p in intermediate_points(points[pidx_a], points[pidx_b], subdiv):
            facet_points.append(len(new_points))
            new_points.append(p)
        facet_points.append(pidx_b)

        for i, p1 in enumerate(facet_points[:-1]):
            p2 = facet_points[i+1]
            new_facets.append((p1, p2))

            if facet_markers is not None:
                new_facet_markers.append(facet_markers[facet_idx])

    if facet_markers is not None:
        return new_points, new_facets, new_facet_markers
    else:
        return new_points, new_facets


def build(mesh_info, verbose=False, refinement_func=None, attributes=False,
        volume_constraints=False, max_volume=None, allow_boundary_steiner=True,
        allow_volume_steiner=True, quality_meshing=True,
        generate_edges=None, generate_faces=False, min_angle=None,
        mesh_order=None, generate_neighbor_lists=False):
    """Triangulate the domain given in `mesh_info'."""
    opts = "pzj"
    if quality_meshing:
        if min_angle is not None:
            opts += "q%f" % min_angle
        else:
            opts += "q"

    if mesh_order is not None:
        opts += "o%d" % mesh_order

    if verbose:
        opts += "VV"
    else:
        opts += "Q"

    if attributes:
        opts += "A"

    if volume_constraints:
        opts += "a"
    if max_volume:
        opts += "a%.20f" % max_volume

    if refinement_func is not None:
        opts += "u"

    if generate_edges is not None:
        from warnings import warn
        warn("generate_edges is deprecated--use generate_faces instead")
        generate_faces = generate_edges
    if generate_neighbor_lists is not None:
        opts += "n"

    if generate_faces:
        opts += "e"

    if not allow_volume_steiner:
        opts += "YY"
        if allow_boundary_steiner:
            raise ValueError("cannot allow boundary Steiner points when volume "
                    "Steiner points are forbidden")
    else:
        if not allow_boundary_steiner:
            opts += "Y"

    # restore "C" locale--otherwise triangle might mis-parse stuff like "a0.01"
    try:
        import locale
    except ImportError:
        have_locale = False
    else:
        have_locale = True
        prev_num_locale = locale.getlocale(locale.LC_NUMERIC)
        locale.setlocale(locale.LC_NUMERIC, "C")

    try:
        mesh = MeshInfo()
        internals.triangulate(opts, mesh_info, mesh, MeshInfo(), refinement_func)
    finally:
        # restore previous locale if we've changed it
        if have_locale:
            locale.setlocale(locale.LC_NUMERIC, prev_num_locale)

    return mesh


def refine(input_p, verbose=False, refinement_func=None,  quality_meshing=True,
        min_angle=None, generate_neighbor_lists=False):
    opts = "razj"

    if quality_meshing:
        if min_angle is not None:
            opts += "q%f" % min_angle
        else:
            opts += "q"

    if len(input_p.faces) != 0:
        opts += "p"
    if verbose:
        opts += "VV"
    else:
        opts += "Q"
    if refinement_func is not None:
        opts += "u"
    if generate_neighbor_lists is not None:
        opts += "n"

    output_p = MeshInfo()
    internals.triangulate(opts, input_p, output_p, MeshInfo(), refinement_func)
    return output_p


def write_gnuplot_mesh(filename, out_p, facets=False):
    gp_file = open(filename, "w")

    if facets:
        segments = out_p.facets
    else:
        segments = out_p.elements

    for points in segments:
        for pt in points:
            gp_file.write("%f %f\n" % tuple(out_p.points[pt]))
        gp_file.write("%f %f\n\n" % tuple(out_p.points[points[0]]))



from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

import numpy
from six.moves import range


class FourDigitsSymmetric:
    def __init__(self, thickness, edge_coeff):
        self.thickness = thickness
        self.edge_coeff = edge_coeff

    def __call__(self, x, side):
        t = self.thickness

        def y_upper(y):
            return y

        def y_lower(y):
            return -y

        def x_upper(x):
            return x

        def x_lower(x):
            return x

        y = t * 5 * (0.2969 * numpy.sqrt(x) + ((((-self.edge_coeff * x +
            0.2843) * x - 0.3516) * x) - 0.126) * x)

        if side == "upper":
            return numpy.array([x_upper(x), y_upper(y)])
        elif side == "lower":
            return numpy.array([x_lower(x), y_lower(y)])
        else:
            raise ValueError("Neither upper nor lower side selected in the call.")


class FourDigitsCambered:
    def __init__(self, thickness, max_camber, max_camber_pos, edge_coeff):
        self.thickness = thickness
        self.max_camber = max_camber
        self.max_camber_pos = max_camber_pos
        self.edge_coeff = edge_coeff

    def __call__(self, x, side):
        t = self.thickness
        m = self.max_camber
        p = self.max_camber_pos

        def y_upper(y_c, y, theta):
            return y_c + y * numpy.cos(theta)

        def y_lower(y_c, y, theta):
            return y_c - y * numpy.cos(theta)

        def x_upper(x, y, theta):
            return x - y * numpy.sin(theta)

        def x_lower(x, y, theta):
            return x + y * numpy.sin(theta)

        y = t * 5 * (0.2969 * numpy.sqrt(x) + ((((-self.edge_coeff * x +
            0.2843) * x - 0.3516) * x) - 0.126) * x)

        if x <= p:
            y_c = m * x / p ** 2 * (2 * p - x)
            theta = numpy.arctan(2 * m / p ** 2 * (p - x))
        else:
            y_c = m * (1 - x) / (1 - p) ** 2 * (1 + x - 2 * p)
            theta = numpy.arctan(m / (1 - p) ** 2 * (2 * p - 2 * x))

        if side == "upper":
            return numpy.array([x_upper(x, y, theta), y_upper(y_c, y, theta)])
        elif side == "lower":
            return numpy.array([x_lower(x, y, theta), y_lower(y_c, y, theta)])
        else:
            raise ValueError("Neither upper nor lower side selected in the call.")


class FiveDigits:
    def __init__(self, thickness, m, k1, edge_coeff):
        self.thickness = thickness
        self.edge_coeff = edge_coeff
        self.m = m
        self.k1 = k1

    def __call__(self, x, side):
        t = self.thickness
        m = self.m
        k1 = self.k1

        def y_upper(y_c, y, theta):
            return y_c + y * numpy.cos(theta)

        def y_lower(y_c, y, theta):
            return y_c - y * numpy.cos(theta)

        def x_upper(x, y, theta):
            return x - y * numpy.sin(theta)

        def x_lower(x, y, theta):
            return x + y * numpy.sin(theta)

        y = t * 5 * (0.2969 * numpy.sqrt(x) + ((((-self.edge_coeff * x +
            0.2843) * x - 0.3516) * x) - 0.126) * x)

        if x <= m:
            y_c = k1 / 6 * x * ((x - 3 * m) * x + m ** 2 * (3 - m))
            theta = numpy.arctan(k1 / 6 * ((3 * x - 6 * m) * x +
                m ** 2 * (3 - m)))
        else:
            y_c = k1 * m ** 3 / 6 * (1 - x)
            theta = numpy.arctan(-k1 * m ** 3 / 6)

        if side == "upper":
            return numpy.array([x_upper(x, y, theta), y_upper(y_c, y, theta)])
        elif side == "lower":
            return numpy.array([x_lower(x, y, theta), y_lower(y_c, y, theta)])
        else:
            raise ValueError("Neither upper nor lower side selected in the call.")


def get_naca_points(naca_digits, number_of_points=100,
        sharp_trailing_edge=True,
        abscissa_map=lambda x: 0.03*x+0.97*x**2,
        verbose=False):
    """
    Return a list of coordinates of NACA 4-digit and 5-digit series
    airfoils.
    """

    if verbose:
        def explain(*s):
            print(" ".join(str(s_i) for s_i in s))
    else:
        def explain(*s):
            pass

    explain("Airfoil: NACA-%s" % naca_digits)

    if sharp_trailing_edge:
        explain("Sharp trailing edge")
        edge_coeff = 0.1036
    else:
        explain("Blunt trailing edge")
        edge_coeff = 0.1015

    raw_abscissae = numpy.linspace(0, 1, number_of_points, endpoint=True)
    abscissae = numpy.empty_like(raw_abscissae)
    for i in range(number_of_points):
        abscissae[i] = abscissa_map(raw_abscissae[i])

    digits_int = int(naca_digits)
    if len(naca_digits) == 4:
        thickness = (digits_int % 100)
        max_camber_pos = (digits_int % 1000) - thickness
        max_camber = (digits_int % 10000) - max_camber_pos - thickness

        thickness = thickness / 1e2
        max_camber_pos = max_camber_pos / 1e3
        max_camber = max_camber / 1e5

        explain("Thickness:", thickness)
        explain("Position of maximum camber:", max_camber_pos)
        explain("Maximum camber:", max_camber)

        if max_camber == 0 and max_camber_pos == 0:
            explain("Symmetric 4-digit airfoil")
            points = FourDigitsSymmetric(thickness, edge_coeff)
        elif max_camber != 0 and max_camber_pos != 0:
            explain("Cambered 4-digit airfoil")
            points = FourDigitsCambered(thickness, max_camber,
                    max_camber_pos, edge_coeff)
        else:
            raise NotImplementedError(
                    "You must decide whether your airfoil shall be cambered or not!")

    elif len(naca_digits) == 5:
        thickness = (digits_int % 100)
        max_camber_pos = (digits_int % 10000) - thickness

        thickness = thickness / 1e2
        max_camber_pos = max_camber_pos / 2e4

        explain("Thickness:", thickness)
        explain("Position of maximum camber:", max_camber_pos)

        identifier = digits_int // 100
        if identifier == 210:
            m = 0.058
            k1 = 361.4
        elif identifier == 220:
            m = 0.126
            k1 = 51.64
        elif identifier == 230:
            m = 0.2025
            k1 = 15.957
        elif identifier == 240:
            m = 0.29
            k1 = 6.643
        elif identifier == 250:
            m = 0.391
            k1 = 3.23
        else:
            raise NotImplementedError("5-digit series only implemented for "
                    "the first three digits in 210, 220, 230, 240, 250!")

        explain("5-digit airfoil")
        points = FiveDigits(thickness, m, k1, edge_coeff)

    else:
        raise NotImplementedError(
                "Only the 4-digit and 5-digit series are implemented!")

    points_upper = numpy.zeros((len(abscissae), 2))
    points_lower = numpy.zeros((len(abscissae), 2))

    for i in range(len(abscissae)):
        points_upper[i] = points(abscissae[i], "upper")
        points_lower[i] = points(abscissae[i], "lower")

    if sharp_trailing_edge:
        return list(points_upper)[1:-1] + list(points_lower[::-1])
    else:
        return list(points_upper)[1:] + list(points_lower[::-1])


def write_points(points, filename):
    file = open(filename, "w")
    for pt in points:
        print("\t".join(repr(p_comp) for p_comp in pt), file=file)


def main():
    from optparse import OptionParser

    parser = OptionParser(usage="%prog AIRFOIL-ID")
    parser.add_option("-o", "--output",
            help="write ouput to FILE", metavar="FILE")
    parser.add_option("-p", "--points", type="int",
            help="generate N points", metavar="N")
    parser.add_option("-s", "--sharp-trailing-edge", action="store_true")
    parser.add_option("-u", "--uniform-distribution", action="store_true")
    parser.add_option("-q", "--quiet",
            action="store_false", dest="verbose", default=True,
            help="Don't print status messages to stdout")

    (options, args) = parser.parse_args()

    if not args:
        parser.print_help()
        return

    if options.points is None:
        options.points = 100

    digits = args[0]
    points = get_naca_points(digits,
            number_of_points=options.points,
            sharp_trailing_edge=options.sharp_trailing_edge,
            uniform_distribution=options.uniform_distribution,
            verbose=options.verbose)

    if options.output is None:
        options.output = "naca-%s.dat" % digits

    print("Output file:", options.output)
    write_points(points, options.output)


if __name__ == "__main__":
    main()
